from sirji_messages import AgentEnum
from sirji_tools.logger import c_logger as logger

from .base import LLMAgentBase


class CodingAgent(LLMAgentBase):
    def __init__(self):
        super().__init__(AgentEnum.CODER, logger)



import os

from sirji_tools.crawler import crawl_urls
from sirji_tools.search import search_for
from sirji_tools.logger import r_logger as logger

from sirji_messages import message_parse, MessageFactory, ActionEnum

from .embeddings.factory import EmbeddingsFactory
from .inferer.factory import InfererFactory


class ResearchAgent:
    def __init__(self, embeddings_type, inferer_type, init_payload={}):
        """Initialize Researcher with specific embeddings and inferer types."""
        logger.info("Initializing researcher...")

        self._embeddings_manager = EmbeddingsFactory.get_instance(
            embeddings_type, init_payload)

        self.init_payload = self._embeddings_manager.init_payload

        self._inferer = InfererFactory.get_instance(
            inferer_type, self.init_payload)

        self._research_folder = os.path.join(self._get_workspace_folder(), '.sirji', self._get_run_id_folder(), "researcher")

        logger.info("Completed initializing researcher")

    def message(self, input_message):
        """Public method to process input messages and dispatch actions."""
        parsed_message = message_parse(input_message)
        action = parsed_message.get("ACTION")

        if action == ActionEnum.TRAIN_USING_SEARCH_TERM.name:
            return self._handle_train_using_search_term(parsed_message)
        elif action == ActionEnum.TRAIN_USING_URL.name:
            return self._handle_train_using_url(parsed_message)
        elif action == ActionEnum.INFER.name:
            return self._handle_infer(parsed_message)
        
    def _get_workspace_folder(self):
        workspace = os.environ.get("SIRJI_WORKSPACE")
        if workspace is None:
            raise ValueError(
                "SIRJI_WORKSPACE is not set as an environment variable")
        return workspace
    
    def _get_run_id_folder(self):
        run_id = os.environ.get("SIRJI_RUN_ID")
        if run_id is None:
            raise ValueError(
                "SIRJI_RUN_ID is not set as an environment variable")
        return run_id

    def _handle_train_using_search_term(self, parsed_message):
        """Private method to handle training using a search term."""
        logger.info(
            f"Training using search term: {parsed_message.get('TERM')}")
        self._search_and_index(parsed_message.get('TERM'))

        return self._generate_message(ActionEnum.TRAINING_OUTPUT, "Training using search term completed successfully"), 0, 0

    def _handle_train_using_url(self, parsed_message):
        """Private method to handle training using a specific URL."""
        logger.info(f"Training using URL: {parsed_message.get('URL')}")
        self._index([parsed_message.get('URL')])

        return self._generate_message(ActionEnum.TRAINING_OUTPUT, "Training using url completed successfully"), 0, 0

    def _handle_infer(self, parsed_message):
        """Private method to handle inference requests."""
        logger.info(f"Infering: {parsed_message.get('DETAILS')}")
        response, prompt_tokens, completion_tokens= self._infer(parsed_message.get('DETAILS'))

        return self._generate_message(ActionEnum.RESPONSE, response), prompt_tokens, completion_tokens

    def _generate_message(self, action_enum, details):
        """Generate standardized messages for responses based on action enum."""
        message_class = MessageFactory[action_enum.name]
        return message_class().generate({"details": details})

    def _index(self, urls):
        """Index given URLs."""
        logger.info("Started indexing the URLs")
        crawl_urls(urls, self._research_folder)
        self.__reindex()
        logger.info("Completed indexing the URLs")

    def _search_and_index(self, query):
        """Search for a query and index resulting URLs."""
        urls = search_for(query)
        self._index(urls)

    def _infer(self, problem_statement):
        """Infer based on the given problem statement and context."""
        retrieved_context = self._embeddings_manager.retrieve_context(
            problem_statement)
        return self._inferer.infer(retrieved_context, problem_statement)

    def __reindex(self):
        """Re-index the research folder recursively. Note: This is treated exceptionally private."""
        logger.info("Recursively indexing the research folder")
        for root, dirs, files in os.walk(self._research_folder):
            for folder in dirs:
                folder_path = os.path.join(root, folder)
                # Improved logging
                logger.info(f"Indexing folder: {folder_path}")

                response = self._embeddings_manager.index(folder_path)
                # Optionally handle the response
        logger.info("Completed recursive indexing of the research folder")



from openai import OpenAI
import os

from sirji_messages import AgentSystemPromptFactory, message_parse, MessageParsingError, MessageValidationError


class SingletonMeta(type):
    """Singleton Meta Class for ensuring one instance creation."""
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


class LLMAgentBase(metaclass=SingletonMeta):
    def __init__(self, agent_enum, logger):
        api_key = os.environ.get("SIRJI_OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "OpenAI API key is not set as an environment variable")

        self.client = OpenAI(api_key=api_key)
        self.agent_enum = agent_enum
        self.logger = logger

    def message(self, input_message, history=[]):
        conversation = self.__prepare_conversation(input_message, history)

        self.logger.info(f"Incoming: \n{input_message}")
        self.logger.info("Calling OpenAI Chat Completions API\n")

        response_message, prompt_tokens, completion_tokens = self.__get_response(conversation)

        return response_message, conversation, prompt_tokens, completion_tokens

    def __prepare_conversation(self, input_message, history):
        conversation = []

        if not history:
            prompt_class = AgentSystemPromptFactory[self.agent_enum.name]
            conversation.append(
                {"role": "system", "content": prompt_class().system_prompt()})
        else:
            conversation = history

        parsed_input_message = message_parse(input_message)
        conversation.append({"role": "user", "content": input_message, "parsed_content": parsed_input_message})

        return conversation

    def __get_response(self, conversation):
        
        retry_llm_count = 0
        response_message = ''
        prompt_tokens = 0
        completion_tokens = 0

        while(True):
            response_message, current_prompt_tokens, current_completion_tokens = self.__call_llm(conversation)
            
            prompt_tokens += current_prompt_tokens
            completion_tokens += current_completion_tokens
            try:
                # Attempt parsing
                parsed_response_message = message_parse(response_message)
                conversation.append({"role": "assistant", "content": response_message, "parsed_content": parsed_response_message})
                break
            except (MessageParsingError, MessageValidationError) as e:
            # Handling both MessageParsingError and MessageValidationError similarly
                self.logger.info("Error while parsing the message.\n")
                retry_llm_count += 1
                if retry_llm_count > 2:
                    raise e
                self.logger.info(f"Requesting LLM to resend the message in correct format.\n")
                conversation.append({"role": "assistant", "content": response_message, "parsed_content": {}})
                conversation.append({"role": "user", "content": "The last message was not as per the allowed message formats. Please resend it with proper formatting."})
            except Exception as e:
                self.logger.info(f"Generic error while parsing message. Error: {e}\n")
                raise e
            
            
        return response_message, prompt_tokens, completion_tokens
    
    def __call_llm(self, conversation):
        history = []

        for message in conversation:
            history.append({"role": message['role'], "content": message['content']})

        chat_completion = self.client.chat.completions.create(
            messages=history,
            model="gpt-4-turbo",
            temperature=0,
            max_tokens=4095,
        )

        response_message = chat_completion.choices[0].message.content

        # Get the total tokens used in the response
        prompt_tokens = chat_completion.usage.prompt_tokens
        completion_tokens = chat_completion.usage.completion_tokens
        self.logger.info(f"Raw response from Chat Completions API: \n{response_message}\n\n\n")

        return response_message, prompt_tokens, completion_tokens



from abc import ABC, abstractmethod


class ResearcherInfererBase(ABC):

    @abstractmethod
    def generate_prompt(self, retrieved_context, problem_statement):
        """
        Generates a prompt that combines the problem statement and the retrieved context.

        :param retrieved_context: The context information retrieved based on embeddings match.
        :param problem_statement: The initial problem statement or query.
        :return: A prompt string that combines both the problem statement and the retrieved context.
        """
        pass

    @abstractmethod
    def infer(self, retrieved_context, problem_statement):
        """
        Infers an answer by calling a chat model using the retrieved context and the problem statement.

        :param retrieved_context: The context information retrieved based on embeddings match.
        :param problem_statement: The initial problem statement or query.
        :return: The model's response based on the combined information of the problem statement and the retrieved context.
        """
        pass



from sirji_messages import AgentEnum
from sirji_tools.logger import p_logger as logger

from .base import LLMAgentBase


class PlanningAgent(LLMAgentBase):
    def __init__(self):
        super().__init__(AgentEnum.PLANNER, logger)



from .openai_assistant import OpenAIAssistantInferer


class InfererFactory:
    @classmethod
    def get_instance(cls, inferer_type, init_payload):
        if inferer_type == "openai_assistant":
            return OpenAIAssistantInferer(init_payload)
        else:
            raise ValueError("Unsupported inferer_type.")
